@@ -2,7 +2,7 @@ from hashlib import md5
 import io
 import os
 import logging
-from .utils import read_block, tokenize
+from .utils import read_block, tokenize, stringify_path
 logger = logging.getLogger('fsspec')
 
 # alternative names for some methods, which get patched to new instances
@@ -36,10 +36,11 @@ class AbstractFileSystem(up):
     from here.
     """
     _singleton = [None]  # will contain the newest instance
-    _cache = None
+    _cache = {}
     cachable = True  # this class can be cached, instances reused
     _cached = False
     blocksize = 2**22
+    sep = "/"
     protocol = 'abstract'
     root_marker = ""  # For some FSs, may require leading '/' or other character
 
@@ -47,24 +48,24 @@ class AbstractFileSystem(up):
         """
         Will reuse existing instance if:
         - cls.cachable is True and
+        - storage_options does not include do_cache=False
         - the token (a hash of args and kwargs by default) exists in the cache
 
         The instance will skip init if instance.cached = True.
         """
-        if cls._cache is None and cls.cachable:
-            # set up instance cache, if using
-            cls._cache = {}
 
         # TODO: defer to a class-specific tokeniser?
+        do_cache = storage_options.pop('do_cache', True)
         token = tokenize(cls, args, storage_options)
-        if cls.cachable and token in cls._cache:
+        if cls.cachable and token in cls._cache and do_cache:
             # check for cached instance
             return cls._cache[token]
         self = object.__new__(cls)
-        self.token = token
+        self._fs_token = token
         if self.cachable:
             # store for caching - can hold memory
             cls._cache[token] = self
+        self.storage_options = storage_options
         return self
 
     def __init__(self, *args, **storage_options):
@@ -100,13 +101,13 @@ class AbstractFileSystem(up):
                     setattr(self, new, getattr(self, old))
 
     def __dask_tokenize__(self):
-        return self.token
+        return self._fs_token
 
     def __hash__(self):
-        return int(self._tok, 16)
+        return int(self._fs_token, 16)
 
     def __eq__(self, other):
-        return self.token == other.token
+        return self._fs_token == other._fs_token
 
     @classmethod
     def clear_instance_cache(cls, remove_singleton=True):
@@ -147,6 +148,7 @@ class AbstractFileSystem(up):
 
         May require FS-specific handling, e.g., for relative paths or links.
         """
+        path = stringify_path(path)
         protos = (cls.protocol, ) if isinstance(
             cls.protocol, str) else cls.protocol
         for protocol in protos:
@@ -212,7 +214,7 @@ class AbstractFileSystem(up):
         """
         pass  # not necessary to implement, may have no cache
 
-    def mkdir(self, path, **kwargs):
+    def mkdir(self, path, create_parents=True, **kwargs):
         """
         Create directory entry at path
 
@@ -223,6 +225,8 @@ class AbstractFileSystem(up):
         ----------
         path: str
             location
+        create_parents: bool
+            if True, this is equivalent to ``makedirs``
         kwargs:
             may be permissions, etc.
         """
@@ -303,7 +307,7 @@ class AbstractFileSystem(up):
                 raise FileNotFoundError(path)
             return files
 
-    def walk(self, path, maxdepth=3):
+    def walk(self, path, maxdepth=None):
         """ Return all files belows path
 
         List all files, recursing into subdirectories; output is iterator-style,
@@ -327,7 +331,7 @@ class AbstractFileSystem(up):
 
         try:
             listing = self.ls(path, True)
-        except FileNotFoundError:
+        except (FileNotFoundError, IOError):
             return [], [], []
 
         for info in listing:
@@ -351,7 +355,7 @@ class AbstractFileSystem(up):
                                      if maxdepth is not None else None):
                     yield res
 
-    def find(self, path, maxdepth=3):
+    def find(self, path, maxdepth=None):
         """List all files below path.
 
         Like posix ``find`` command without conditions
@@ -367,9 +371,9 @@ class AbstractFileSystem(up):
             # walk works on directories, but find should also return [path]
             # when path happens to be a file
             out.append(path)
-        return out
+        return sorted(out)
 
-    def du(self, path, total=True, maxdepth=4):
+    def du(self, path, total=True, maxdepth=None):
         """Space used by files within a path
 
         Parameters
@@ -420,7 +424,7 @@ class AbstractFileSystem(up):
             elif self.exists(path):
                 return [path]
             else:
-                raise FileNotFoundError(path)
+                return []  # glob of non-existent returns empty
         elif '/' in path[:ind]:
             ind2 = path[:ind].rindex('/')
             root = path[:ind2 + 1]
@@ -429,12 +433,11 @@ class AbstractFileSystem(up):
             root = ''
             depth = 20 if "**" in path else 1
         allpaths = self.find(root, maxdepth=depth)
-        pattern = re.compile("^" + path.replace('.', r'\.')
-                             .replace('//', '/')
-                             .rstrip('/')
-                             .replace('**', '.+')
-                             .replace('*', '[^/]*')
-                             .replace('?', '.') + "$")
+        pattern = "^" + path.replace('.', r'\.').replace('//', '/').rstrip(
+            '/').replace('?', '.') + "$"
+        pattern = re.sub('[*]{2}', '=PLACEHOLDER=', pattern)
+        pattern = re.sub('[*]', '[^/]*', pattern)
+        pattern = re.compile(pattern.replace("=PLACEHOLDER=", '.*'))
         out = {p for p in allpaths
                if pattern.match(p.replace('//', '/').rstrip('/'))}
         return list(sorted(out))
@@ -520,14 +523,7 @@ class AbstractFileSystem(up):
         rpath = self._strip_protocol(rpath)
         if recursive:
             rpaths = self.find(rpath)
-            rootdir = os.path.basename(rpath.rstrip('/'))
-            if os.path.isdir(lpath):
-                # copy rpath inside lpath directory
-                lpath2 = os.path.join(lpath, rootdir)
-            else:
-                # copy rpath as lpath directory
-                lpath2 = lpath
-            lpaths = [os.path.join(lpath2, path[len(rpath):].lstrip('/'))
+            lpaths = [os.path.join(lpath, path[len(rpath):].lstrip('/'))
                       for path in rpaths]
             for lpath in lpaths:
                 dirname = os.path.dirname(lpath)
@@ -537,7 +533,6 @@ class AbstractFileSystem(up):
             rpaths = [rpath]
             lpaths = [lpath]
         for lpath, rpath in zip(lpaths, rpaths):
-            print(lpath, rpath)
             with self.open(rpath, 'rb') as f1:
                 with open(lpath, 'wb') as f2:
                     data = True
@@ -619,14 +614,15 @@ class AbstractFileSystem(up):
                 out = self.walk(p, maxdepth=maxdepth)
                 for pa, _, files in reversed(list(out)):
                     for name in files:
-                        self.rm('/'.join([pa, name]))
+                        fn = '/'.join([pa, name]) if pa else name
+                        self.rm(fn)
                     self.rmdir(pa)
             else:
                 self._rm(p)
 
     @classmethod
     def _parent(cls, path):
-        path = path.rstrip('/').lstrip('/')
+        path = cls._strip_protocol(path.rstrip('/'))
         if '/' in path:
             return cls.root_marker + path.rsplit('/', 1)[0]
         else:
@@ -844,7 +840,7 @@ class AbstractBufferedFile(io.IOBase):
         self.fs = fs
         self.mode = mode
         self.blocksize = (self.DEFAULT_BLOCK_SIZE
-                          if block_size == 'default' else block_size)
+                          if block_size in ['default', None] else block_size)
         self.loc = 0
         self.autocommit = autocommit
         self.end = None
@@ -876,7 +872,10 @@ class AbstractBufferedFile(io.IOBase):
         self._closed = c
 
     def __hash__(self):
-        return self.fs.checksum(self.path)
+        if 'w' in self.mode:
+            return id(self)
+        else:
+            return int(tokenize(self.details), 16)
 
     def __eq__(self, other):
         """Files are equal if they have the same checksum, only in read mode"""
@@ -891,7 +890,10 @@ class AbstractBufferedFile(io.IOBase):
 
     def info(self):
         """ File information about this path """
-        return self.details  # error in write mode
+        if 'r' in self.mode:
+            return self.details
+        else:
+            raise ValueError('Info not available while writing')
 
     def tell(self):
         """ Current file location """
@@ -1018,11 +1020,11 @@ class AbstractBufferedFile(io.IOBase):
         length : int (-1)
             Number of bytes to read; if <0, all remaining bytes.
         """
-        length = int(length)
+        length = -1 if length is None else int(length)
         if self.mode != 'rb':
             raise ValueError('File not in read mode')
         if length < 0:
-            length = self.size
+            length = self.size - self.loc
         if self.closed:
             raise ValueError('I/O operation on closed file.')
         logger.debug("%s read: %i - %i" % (self, self.loc, self.loc + length))
@@ -1090,7 +1092,7 @@ class AbstractBufferedFile(io.IOBase):
         lines = data.split(b'\n')
         out = [l + b'\n' for l in lines[:-1]]
         if data.endswith(b'\n'):
-            return out + [lines[-1] + b'\n']
+            return out
         else:
             return out + [lines[-1]]
         # return list(self)  ???
@@ -1113,8 +1115,9 @@ class AbstractBufferedFile(io.IOBase):
             else:
                 assert self.buffer.tell() == 0
 
-            self.fs.invalidate_cache(self.path)
-            self.fs.invalidate_cache(self.fs._parent(self.path))
+            if self.fs is not None:
+                self.fs.invalidate_cache(self.path)
+                self.fs.invalidate_cache(self.fs._parent(self.path))
 
         self.closed = True
 
